#ifndef SYSTEM_DISPLAY_H
#define SYSTEM_DISPLAY_H

#include "math/vec.h"

class PDisplay {
public:
	virtual void update() = 0;
	virtual void resize(int w, int h) = 0;
	virtual const vec2i &getSize() const = 0;
	virtual bool sizeHasChanged() = 0;
};

class Display : public PDisplay {
protected:

	//current dimensions
	vec2i	size;		//int width, height;
	//last update's dimensions
	vec2i	lastsize;	//int lastWidth, lastHeight;

	//set through 'resize', updated upon 'update'
	vec2i	newsize;	//int newWidth, newHeight;
	
public:
	Display() {reset();}

	void reset();

	virtual void update();

	//called by system upon resize.
	//writes changes to newWidth/Height.
	//updates upon 'update' so old res is stored
	//this way referencing applications can detect a resize, and can adjust accordingly
	virtual void resize(int w, int h);
	virtual const vec2i &getSize() const { return size; }
	virtual bool sizeHasChanged() { return size != lastsize; }
};

class IDisplay {
	PDisplay *imp;
public:
	IDisplay();
	void setImp(PDisplay *_imp);
	virtual void update();
	virtual void resize(int w, int h);
	virtual const vec2i &getSize() const;
	virtual bool sizeHasChanged();
};

#endif	//SYSTEM_DISPLAY_H
