#ifndef SYSTEM_EXCEPTION_H
#define SYSTEM_EXCEPTION_H

//TODO - make virtual somehow

//exception testing
//of dynamic content
//across multiple methods
#include <iostream>
#include <string>
#include <sstream>

//The only way I could think to retain the operator<<
//without having the return type perpetually skewed to the parent class type
#ifndef WIN32
#define THROWABLE_ADD_SHIFT_OPERATOR(class)	\
template<typename T> inline class &operator<<(const T &stuff) {	\
	s << stuff;	\
	return *this;	\
}
#else
#define THROWABLE_ADD_SHIFT_OPERATOR(class)	\
template<typename T> inline class &operator<<(const T &stuff) {	\
	s << stuff;	\
	return *this;	\
}	\
inline class &operator<<(const char *stuff) {	\
	s << (stuff ? stuff : "(null)");	\
	return *this;	\
}
#endif

class Throwable {
protected:
	std::ostringstream s;
public:
	Throwable() { }
	Throwable(const Throwable &t) { s << t.s.str(); }
	
	std::string toString() const { return s.str(); }
	
	THROWABLE_ADD_SHIFT_OPERATOR(Throwable)
};

inline std::ostream &operator<<(std::ostream &s, const Throwable &t) {
	return s << t.toString();
}

//common throwable exception
class Exception : public Throwable {
	typedef Throwable super;
public:
	Exception() : super() {}
	Exception(const Throwable &t) : super(t) {}
	THROWABLE_ADD_SHIFT_OPERATOR(Exception)
};

//error
class Error : public Throwable {
	typedef Throwable super;
public:
	Error() : super() {}
	Error(const Throwable &t) : super(t) {}
	THROWABLE_ADD_SHIFT_OPERATOR(Error)
};

#endif

