#ifndef SYSTEM_KEYBOARD_H
#define SYSTEM_KEYBOARD_H

#include "system/defs.h"	//needed for keystate interpretation

//public

//i'm gonna remap all keys from SDL to my own junk
//that way no one can have the better
//and i'll be fully platform independent
class Key {
public:
	//keyboard mapping ...
	//soon to replace SDL
	enum Key_t {
		BACKSPACE = 8,
		TAB,
		RETURN = 13,
		ESC = 27,
		SPACE = 32,
		//ascii 0-127 is mapped directly across
		//at least shift-less characters are ...
		//at some time i should fix thatnt.h
		DEL = 127,

		LSHIFT, RSHIFT,
		LCTRL, RCTRL,
		LALT, RALT,
		LMETA, RMETA,	//apple key for apples, windows key for windows?

		F1,F2,F3,F4,
		F5,F6,F7,F8,
		F9,F10,F11,F12,
		F13,F14,F15,F16,

		UP, DOWN, LEFT, RIGHT,
		PAGE_UP,
		PAGE_DOWN,
		HOME,
		END,
		INSERT,
		//delete is 127

		KP0,KP1,KP2,KP3,
		KP4,KP5,KP6,KP7,
		KP8,KP9,
		
		CAPSLOCK,

		COUNT
	};

	static const char *name[COUNT+1];
	static const char *nameEnd;	//ptr used to denote the end of the name array - for validity testing
	
	//this basically works on ascii codes
	//since thats what chars 0-127 are mapped to
	//but it could potentially work on others ...
	static int removeShift[COUNT];
	static int addShift[COUNT];
};

class PKeyboardHandler {
public:
	//the state of a key - PRESSED, DOWN, RELEASED, UP
	virtual char getKeyState(int key) = 0;

	virtual void reset() = 0;
	virtual void keyUp(Key::Key_t key) = 0;
	virtual void keyDown(Key::Key_t key) = 0;
	virtual void preUpdate() = 0;
	virtual void postUpdate() = 0;
	virtual int getKeyDownQueueSize() = 0;
	virtual int getKeyDownQueue(int i) = 0;

	//console requests
	virtual PKeyboardHandler *clone() = 0;	//returns a newly allocated copy
	virtual void copy(PKeyboardHandler *from) = 0;	//copies from the specified
};

class IKeyboardHandler : public PKeyboardHandler {
	PKeyboardHandler *imp;
public:
	IKeyboardHandler();
	void setImp(PKeyboardHandler *_imp);
	PKeyboardHandler *getImp();
	
	virtual void reset();
	virtual char getKeyState(int key);
	virtual int getKeyDownQueueSize();
	virtual int getKeyDownQueue(int i);
	virtual void keyUp(Key::Key_t key);
	virtual void keyDown(Key::Key_t key);
	virtual void preUpdate();
	virtual void postUpdate();
	
	virtual PKeyboardHandler *clone();	//returns a newly allocated copy
	virtual void copy(PKeyboardHandler *from);	//copies from the specified

	//operator[] access
	class KeyStateAccessor {
		IKeyboardHandler &h;
	public:
		KeyStateAccessor(IKeyboardHandler &_h) : h(_h) {}
		char operator[](int key) { return h.getKeyState(key); }
	};
	KeyStateAccessor keystate;
};

//private

class KeyboardHandler : public PKeyboardHandler {
public:
	KeyboardHandler() { reset(); }

	//called upon init and before the window loses focus
	//sets all variables to default value
	virtual void reset();

	//call when a key is pressed
	virtual void keyUp(Key::Key_t key);
	//call when a key is depressed
	virtual void keyDown(Key::Key_t key);
	//for intermediate.  use this last-case, cuz it bypasses the keydownqueue
	void keySet(Key::Key_t key, bool state);

	//call to calculate the keystate array and copy the current keypress over the last
	virtual void preUpdate();
	//call to clear the keydown queue
	virtual void postUpdate();

	//virtual access for the interface
	virtual char getKeyState(int key);

	virtual int getKeyDownQueueSize();
	virtual int getKeyDownQueue(int i);

	virtual PKeyboardHandler *clone();	//returns a newly allocated copy
	virtual void copy(PKeyboardHandler *from);	//copies from the specified

protected:

	//the state of a key - PRESSED, DOWN, RELEASED, UP
	char keystate[Key::COUNT];

	//store all the keystrokes that have been pressed since the last postUpdate
	int keydownqueue[32];
	int firstkeydown, lastkeydown;

	//whether a key is up or down (0 or 1, respectively)
	//TODO - bitvector this, for size's sake?
	char keypress[Key::COUNT];
	char keypress_last[Key::COUNT];
};

#endif