#include <string.h>	//for memcpy and memset
#include <assert.h>
#include <stdio.h>
#include "defs/defs.h"	//numberof
#include "system/keyboard.h"

const char *Key::nameEnd = "";	//make sure this points to some const string

//true the world over.  consider putting this in system base ...
//or make a sep 'key.cpp/h' files to put there...
const char *Key::name[COUNT+1] = {
	"",	"",	"",	"",	"",	"",	"",	"",
	"BACKSPACE",	"TAB",	"",	"",	"",	"RETURN",	"",	"",
	"",	"",	"",	"",	"",	"",	"",	"",
	"",	"",	"",	"ESC",	"",	"",	"",	"",
	"SPACE",
	//and for windows shift code virtual key reasons I'm exluding these shifted keys ...
	"",	"",	"",	"",	"",	"",	//"!", "\"", "#",	"$",	"%",	"&",	
	"'",
	"",	"",	"",	"",	//"(",	")",	"*",	"+",
	",",	"-",	".",	"/",
	"0",	"1",	"2",	"3",	"4",	"5",	"6",	"7",
	"8",	"9",	
	"",	//":",
	";",
	"",	//"<",
	"=",
	"",	"", //">",	"?",
	"",	//"@",
	//this is where upper case letter pick up
	//mind you its PC to chart them as they are
	//but since their keys are unreachable ... I'd rather pretend they dont exist
	//that way case-insensitive keyname searches wont reference these unusable keys first
	"",	"",	"",	"",	"",	"",	"",		//	"A",	"B",	"C",	"D",	"E",	"F",	"G",
	"",	"",	"",	"",	"",	"",	"",	"",	//	"H",	"I",	"J",	"K",	"L",	"M",	"N",	"O",
	"",	"",	"",	"",	"",	"",	"",	"",	//	"P",	"Q",	"R",	"S",	"T",	"U",	"V",	"W",
	"",	"",	"",						//	"X",	"Y",	"Z",	
	"[",	"\\",	"]",	
	"",	"",	//"^",	"_",
	"`",	"a",	"b",	"c",	"d",	"e",	"f",	"g",
	"h",	"i",	"j",	"k",	"l",	"m",	"n",	"o",
	"p",	"q",	"r",	"s",	"t",	"u",	"v",	"w",
	"x",	"y",	"z",	
	"", "", "", "", //"{",	"|",	"}",	"~",
	
	"DEL",
	"LSHIFT",	"RSHIFT",
	"LCTRL",	"RCTRL",
	"LALT",		"RALT",
	"LMETA",	"RMETA",

	"F1","F2","F3","F4",
	"F5","F6","F7","F8",
	"F9","F10","F11","F12",
	"F13","F14","F15","F16",

	"UP",	"DOWN",	"LEFT",	"RIGHT",
	"PAGE_UP",	"PAGE_DOWN",
	"HOME",	"END",
	"INSERT",

	"KP0","KP1","KP2","KP3",
	"KP4","KP5","KP6","KP7",
	"KP8","KP9",
	
	"CAPSLOCK",
	
	nameEnd	//for validity testing
};

int Key::removeShift[COUNT]
#if 0
 = {
	0,0,0,0,0,0,0,0,			//"",	"",	"",	"",	"",	"",	"",	"",
	0,0,0,0,0,0,0,0,			//"BACKSPACE",	"TAB",	"",	"",	"",	"RETURN",	"",	"",
	0,0,0,0,0,0,0,0,			//"",	"",	"",	"",	"",	"",	"",	"",
	0,0,0,0,0,0,0,0,			//"",	"",	"",	"ESC",	"",	"",	"",	"",
	0,							//"SPACE",
	//and for windows shift code virtual key reasons I'm exluding these shifted keys ...
	'1', '\'', '3', '4', '5', '7',	//"!", "\"", "#",	"$",	"%",	"&",	
	'\'',
	'9', '0', '8', '=',	//"(",	")",	"*",	"+",
	',',	'-',	'.',	'/',
	'0',	'1',	'2',	'3',	'4',	'5',	'6',	'7',
	'8',	'9',	
	';',	//":",
	';',
	',',	//"<",
	'=',
	'.', '/', //">",	"?",
	'2',	//"@",
	//this is where upper case letter pick up
	//mind you its PC to chart them as they are
	//but since their keys are unreachable ... I'd rather pretend they dont exist
	//that way case-insensitive keyname searches wont reference these unusable keys first
	'a', 'b', 'c', 'd', 'e', 'f', 'g',		//	"A",	"B",	"C",	"D",	"E",	"F",	"G",
	'h', 'i', 'j', 'k', 'l', 'm', 'n', 'o',	//	"H",	"I",	"J",	"K",	"L",	"M",	"N",	"O",
	'p', 'q', 'r', 's', 't', 'u', 'v', 'w',	//	"P",	"Q",	"R",	"S",	"T",	"U",	"V",	"W",
	'x', 'y', 'z',						//	"X",	"Y",	"Z",	
	'[', '\\', ']',	
	'6', '-',	//"^",	"_",
	'`',	'a',	'b',	'c',	'd',	'e',	'f',	'g',
	'h',	'i',	'j',	'k',	'l',	'm',	'n',	'o',
	'p',	'q',	'r',	's',	't',	'u',	'v',	'w',
	'x',	'y',	'z',
	'[', '\\', ']', '`', //"{",	"|",	"}",	"~",
}
#endif
;
int Key::addShift[Key::COUNT];

void KeyboardHandler::reset() {
	memset(keypress, 0, sizeof(keypress));
	memset(keypress_last, 0, sizeof(keypress));
	memset(keystate, 0, sizeof(keystate));
	memset(keydownqueue, 0, sizeof(keydownqueue));

	//only need to run this once
	//but i'm lazy so it goes here

	for (int i = 0; i < Key::COUNT; i++) {
		Key::removeShift[i] = i;
		Key::addShift[i] = i;
	}
#define SHIFTBRIDGE(x,y)	Key::removeShift[y] = x; Key::addShift[x] = y;
#define SHIFTBRIDGERANGE(x,y,z)	\
	for (int i = x; i <= y; i++) {	\
		Key::removeShift[i-x+z] = i;	\
		Key::addShift[i] = i-x+z;	\
	}
	SHIFTBRIDGE('1','!') SHIFTBRIDGE('\'','\"') SHIFTBRIDGE('3','#') SHIFTBRIDGE('4','$')
	SHIFTBRIDGE('5','%') SHIFTBRIDGE('7','&') SHIFTBRIDGE('9','(') SHIFTBRIDGE('0',')')
	SHIFTBRIDGE('8','*') SHIFTBRIDGE('=','+') SHIFTBRIDGE(';',':') SHIFTBRIDGE(',','<')
	SHIFTBRIDGE('.','>') SHIFTBRIDGE('/','?') SHIFTBRIDGE('2','@') SHIFTBRIDGE('6','^')
	SHIFTBRIDGE('-','_') SHIFTBRIDGE('[','{') SHIFTBRIDGE('\\','|') SHIFTBRIDGE(']','}')
	SHIFTBRIDGE('`','~')
	SHIFTBRIDGERANGE('a','z','A')

	assert(Key::removeShift['A'] == 'a');

	firstkeydown = lastkeydown = 0;
}

void KeyboardHandler::keyUp(Key::Key_t key) {
//	assert(key >= 0 && key < numberof(keypress));
	if (!(key >= 0 && key < numberof(keypress))) {
		fprintf(stderr, "KeyboardHandler::keyUp got an unknown key %d\n", key);
		return;
	}
	keypress[key] = 0;
}

void KeyboardHandler::keyDown(Key::Key_t key) {
//	assert(key >= 0 && key < numberof(keypress));
	if (!(key >= 0 && key < numberof(keypress))) {
		fprintf(stderr, "KeyboardHandler::keyUp got an unknown key %d\n", key);
		return;
	}

	//record in the keydown queue
	keydownqueue[lastkeydown] = key;
	//calc the new queue end location
	int nextkeydown = (lastkeydown + 1) % numberof(keydownqueue);
	//if it doesnt surpass the beginning, then increment
	if (nextkeydown != firstkeydown) lastkeydown = nextkeydown;

	//set the current key state to 'down'
	keypress[key] = 1;
}

void KeyboardHandler::keySet(Key::Key_t key, bool state) {
//	assert(key >= 0 && key < numberof(keypress));
	if (!(key >= 0 && key < numberof(keypress))) {
		fprintf(stderr, "KeyboardHandler::keyUp got an unknown key %d\n", key);
		return;
	}

	keypress[key] = state;
}

char KeyboardHandler::getKeyState(int key) {
	return keystate[key];
}

int KeyboardHandler::getKeyDownQueueSize() {
	return (lastkeydown - firstkeydown + numberof(keydownqueue)) % numberof(keydownqueue);
}
int KeyboardHandler::getKeyDownQueue(int i) {
	return keydownqueue[(firstkeydown + i) % numberof(keydownqueue)];
}

void KeyboardHandler::preUpdate() {

	// *************** INTRO - calc the keystat array via the keypress and keypress_last arrays

	for (int x = 0; x < numberof(keypress); x++) {
		if (keypress[x]) {	//key down
			if (keypress_last[x])		//it is pressed and has been presssed
				keystate[x] = BUTTON_STATE_DOWN;
			else					//is being pressed but has not been before
				keystate[x] = BUTTON_STATE_PRESS;
		} else {			//key up
			if (keypress_last[x])		//is being released and was being pressed
				keystate[x] = BUTTON_STATE_RELEASE;
			else					//is up and has been up
				keystate[x] = BUTTON_STATE_UP;
		}
	}

	// *************** ENCORE - copy the keypress map to the keypress_last map

	memcpy(keypress_last, keypress, sizeof(keypress));
}

//this is a post-update so all else between has a chance to view the current key queue (ie the console)
void KeyboardHandler::postUpdate() {
	firstkeydown = lastkeydown;
}

PKeyboardHandler *KeyboardHandler::clone() {
	KeyboardHandler *h = new KeyboardHandler();
	h->copy(this);
	return h;
}

void KeyboardHandler::copy(PKeyboardHandler *_from) {
	KeyboardHandler *from = dynamic_cast<KeyboardHandler*>(_from);
	assert(from);
	//all members...
	memcpy(keystate, from->keystate, sizeof(keystate));
	memcpy(keydownqueue, from->keydownqueue, sizeof(keydownqueue));
	firstkeydown = from->firstkeydown;
	lastkeydown = from->lastkeydown;
	memcpy(keypress, from->keypress, sizeof(keypress));
	memcpy(keypress_last, from->keypress_last, sizeof(keypress_last));
}

IKeyboardHandler::IKeyboardHandler() : imp(NULL), keystate(*this) {}
void IKeyboardHandler::setImp(PKeyboardHandler *_imp) { imp = _imp; }
PKeyboardHandler *IKeyboardHandler::getImp() { return imp; }

void IKeyboardHandler::reset() {
	assert(imp);
	return imp->reset();
}
char IKeyboardHandler::getKeyState(int key) {
	assert(imp);
	return imp->getKeyState(key);
}

void IKeyboardHandler::keyUp(Key::Key_t key) {
	assert(imp);
	imp->keyUp(key);
}

void IKeyboardHandler::keyDown(Key::Key_t key) {
	assert(imp);
	imp->keyDown(key);
}

void IKeyboardHandler::preUpdate() {
	assert(imp);
	imp->preUpdate();
}

void IKeyboardHandler::postUpdate() {
	assert(imp);
	imp->postUpdate();
}
int IKeyboardHandler::getKeyDownQueueSize() {
	assert(imp);
	return imp->getKeyDownQueueSize();
}
int IKeyboardHandler::getKeyDownQueue(int i) {
	assert(imp);
	return imp->getKeyDownQueue(i);
}
PKeyboardHandler *IKeyboardHandler::clone() {
	assert(imp);
	return imp->clone();
}
void IKeyboardHandler::copy(PKeyboardHandler *k) {
	assert(imp);
	imp->copy(k);
}

