#ifndef WIN32

#include <assert.h>
#include <pthread.h>

#include "system/mutex.h"

class LinuxMutex : public IMutex {
protected:
	pthread_mutex_t mutex;
	bool initd;
public:
	LinuxMutex(bool initOnCtor = true);
	virtual ~LinuxMutex();
	
	virtual bool init();
	virtual void shutdown();
	
	virtual void lock();
	virtual void unlock();
};

LinuxMutex::LinuxMutex(bool initOnCtor) : initd(false) {
	if (initOnCtor) assert(init());
}

LinuxMutex::~LinuxMutex() { shutdown(); }

bool LinuxMutex::init() {
	if (pthread_mutex_init(&mutex, NULL)) return false;
	initd = true;
	return true;
}
void LinuxMutex::shutdown() {
	if (initd) pthread_mutex_destroy(&mutex);
	initd = false;
}

void LinuxMutex::lock() { pthread_mutex_lock(&mutex); }
void LinuxMutex::unlock() { pthread_mutex_unlock(&mutex); }

IMutex *MutexFactory::create(bool init) {
	return new LinuxMutex(init);
}

#endif	//WIN32
