#include <stdlib.h>
#include <iostream>
using namespace std;

#include "system/memory.h"

//two things here:
//1) console isnt up yet
//2) memory is in the system, which depends on console, so it cant ref back the console
//#define USE_MEMORY_WITH_CONSOLE

#define MPTR_IDENT		0xA0B1C2D3

typedef struct mptr_s {
	struct mptr_s	*prev, *next;
	size_t			size;
	int				ident;
	const char		*name;
} mptr_t;

static mptr_t *mptr_first, *mptr_last;

static size_t total_size = 0;
static int total_blocks = 0;

bool memValid(void *data) {
	mptr_t *p = ((mptr_t *)data)-1;
	return p->ident == MPTR_IDENT;
}

#ifdef USE_MEMORY_WITH_CONSOLE
void CFunc_Mem(int argc, char **argv) {
	if (argc > 1 && !strcmp(argv[1], "info")) {
		for (mptr_t *p = mptr_first; p; p=p->next) {
			if (!memValid(p+1)) con.print("[bad]");
			con.print("  %d - %s\n", p->size, p->name);
		}
	}
	con.print("  total memory size: %u\n", total_size);
	con.print("  total memory blocks: %d\n", total_blocks);
}
static CVarw cv_mem(0, "mem", CFunc_Mem);
#endif //USE_MEMORY_WITH_CONSOLE

void memFree(void *data) {
	mptr_t *p = ((mptr_t *)data)-1;

	if (!memValid(data)) {
		cout << "memFree: found a bad ID: " << p->ident << " should be " << MPTR_IDENT << endl;
	}

	total_size -= p->size;
	total_blocks--;
	
	//unlink
	if (p->prev) {
		p->prev->next = p->next;
	} else {	//no prev == is first
		mptr_first = p->next;
	}
	if (p->next) {
		p->next->prev = p->prev;
	} else {	//no next == islast
		mptr_last = p->prev;
	}

	free(p);
}

//addlast
static void MPtr_Link(mptr_t *p) {
	if (mptr_last) {
		mptr_last->next = p;
	} else {
		mptr_first = p;
	}
	p->prev = mptr_last;
	p->next = NULL;
	mptr_last = p;
}

void *memAllocNamed(size_t size, const char *name) {
	size += sizeof(mptr_t);
	mptr_t *p = (mptr_t *)malloc(size);
	if (!p) {
		cout << "failed to alloc " << (int)size << " bytes of type " << name << endl;
		return NULL;
	}

	p->size = size;
	p->name = name;
	p->ident = MPTR_IDENT;

	MPtr_Link(p);

	total_size += size;
	total_blocks++;

	//ahh, so the trick is to add in a lil extra for the pointer link
	return (void *)(p+1);
}

void *memAlloc(size_t size) {
	return memAllocNamed(size, "Generic");
}

static void memSystemClear() {
	mptr_first = mptr_last = NULL;
	total_size = 0;
	total_blocks = 0;
}

void memSystemInit() {
	memSystemClear();
}

void memSystemShutdown() {
	mptr_t *p, *next;
	for (p = mptr_first; p; p=next) {
		next = p->next;	//store the next pointer before we destroy 'p'
		memFree(p+1);	//call free on the data, not the important stuff
	}

	memSystemClear();	//wipe it all clean
}