#include <memory.h>
#include "system/singletons.h"
#include "system/mouse.h"

void MouseHandler::reset() {
	pos = 0;
	move = 0;
	posf = 0.f;
	movef = 0.f;
	dz = 0;
	dzLast = 0;
	newpos = 0;
	memset(buttonPress, 0, sizeof(buttonPress));
	memset(buttonPressLast, 0, sizeof(buttonPressLast));
}

void MouseHandler::update() {
	
	//flip the y coordinate
	vec2i flipNewPos(newpos);
	flipNewPos.y = disp.getSize().y - flipNewPos.y - 1;

	move = flipNewPos - pos;

	pos = flipNewPos;

	posf.x = (float)pos.x / (float)disp.getSize().x;
	posf.y = (float)pos.y / (float)disp.getSize().y;

	movef.x = (float)move.x / (float)disp.getSize().x;
	movef.y = (float)move.y / (float)disp.getSize().y;

	dz = dzLast;
	dzLast = 0;

	//THIS CODE MIRRORS THE KEYBOARD!
	//MERGE THE TWO!
	for (int i = 0; i < numberof(buttonPress); i++) {
		if (buttonPress[i])	{	//key down
			if (buttonPressLast[i])		//it is pressed and has been presssed
				buttonState[i] = BUTTON_STATE_DOWN;
			else					//is being pressed but has not been before
				buttonState[i] = BUTTON_STATE_PRESS;
		} else {			//key up
			if (buttonPressLast[i])		//is being released and was being pressed
				buttonState[i] = BUTTON_STATE_RELEASE;
			else					//is up and has been up
				buttonState[i] = BUTTON_STATE_UP;
		}
	}

	//update new state
	memcpy(buttonPressLast, buttonPress, sizeof(buttonPress));
}


//this function can only be called by main_display
//therefore in order for this to be called, lets assume main_display is up and running
void MouseHandler::setButtonPress(Button_t button, bool down) {
	if(!(button >= 0 && button < numberof(buttonPress))) {
		fprintf(stderr, "MouseHandler::setButtonPress got an unknown button %d\n", button);
		return;
	}
	buttonPress[button] = down;
}

void MouseHandler::setCursorPos(int x, int y) {
	newpos.set(x,y);
}

void MouseHandler::accumDZ(short addDZ) {
	dzLast += addDZ;
}

IMouseHandler::IMouseHandler() : imp(NULL) {}
void IMouseHandler::setImp(PMouseHandler *_imp) { imp = _imp; }
void IMouseHandler::update() {
	assert(imp);
	imp->update();
}
void IMouseHandler::setButtonPress(Button_t button, bool down) {
	assert(imp);
	imp->setButtonPress(button, down);
}
void IMouseHandler::setCursorPos(int x, int y) {
	assert(imp);
	imp->setCursorPos(x,y);
}
void IMouseHandler::accumDZ(short addDZ) {
	assert(imp);
	imp->accumDZ(addDZ);
}
int IMouseHandler::getDZ() const {
	assert(imp);
	return imp->getDZ();
}
const vec2i &IMouseHandler::getPos() const {
	assert(imp);
	return imp->getPos();
}
const vec2i &IMouseHandler::getMove() const {
	assert(imp);
	return imp->getMove();
}
const vec2f &IMouseHandler::getFPos() const {
	assert(imp);
	return imp->getFPos();
}
const vec2f &IMouseHandler::getFMove() const {
	assert(imp);
	return imp->getFMove();
}
const char IMouseHandler::getButtonState(int button) const {
	assert(imp);
	return imp->getButtonState(button);
}
