#include "system/thread.h"
#include "defs/defs.h"

#ifdef WIN32
#define THREAD_PROC_TYPE	DWORD
#else
#include <pthread.h>	//isnt this included through thread.h below?
#include <stdio.h>
#define THREAD_PROC_TYPE	void *
#endif

static THREAD_PROC_TYPE ThreadProc(void *arg) {
	Thread *thread = (Thread *)arg;
	int result = thread->run();
	thread->thread = 0;
	return (THREAD_PROC_TYPE)result;
}

#ifdef WIN32

bool Thread::init() {
	unsigned long threadID;	//what is this used for? thread identifier?
	thread = CreateThread(
		NULL, 											// pointer to thread security attributes 
		0,												// initial thread dataStack size, in bytes 
		(LPTHREAD_START_ROUTINE)ThreadProc,			// pointer to thread function 
		this,											// argument for new thread 
		0,												// creation flags 
		&threadID);										// pointer to returned thread identifier 
	if (!thread) return false;

	return true;
}

void Thread::setPriority(Thread::Priority pr) {
	SetThreadPriority(thread, (int)pr);	//are you sure the priorities map so directly?
}

bool Thread::wait() {
	return WaitForSingleObject(thread, INFINITE) != WAIT_FAILED;
}

#else

bool Thread::init() {
	if (pthread_create(
		&thread,	//thread
		NULL,	//attr
		ThreadProc,	//start_routine,
		this))	//arg
	{
		return false;
	}

	return true;
}

#ifndef PTHREAD_RT_PRIORITY
#define PTHREAD_RT_PRIORITY 63
#endif
#ifndef PTHREAD_MAX_PRIORITY
#define PTHREAD_MAX_PRIORITY 31
#endif
#ifndef PTHREAD_MIN_PRIORITY
#define PTHREAD_MIN_PRIORITY 0
#endif

#define PRIORITY(x)		((int)((double)PTHREAD_MAX_PRIORITY*x + (double)PTHREAD_MIN_PRIORITY*(1.0-x)))

static int pthreadPriorityMapping[Thread::PRIORITY_COUNT] = {
		PTHREAD_RT_PRIORITY,	//PRIORITY_TIME_CRITICAL,
		PRIORITY(1.),	//PRIORITY_HIGHEST,
		PRIORITY(.75),	//PRIORITY_ABOVE_NORMAL,
		PRIORITY(.5),	//PRIORITY_NORMAL,
		PRIORITY(.375),	//PRIORITY_BELOW_NORMAL,
		PRIORITY(.25),	//PRIORITY_LOWEST,
		PRIORITY(.125),	//PRIORITY_ABOVE_IDLE,
		PRIORITY(0.),	//PRIORITY_IDLE,
};

void Thread::setPriority(Thread::Priority pr) {
	//how do you set priority in posix threads?
	if (pr < 0) pr = (Thread::Priority)0;
	if (pr >= numberof(pthreadPriorityMapping)) {
		pr = (Thread::Priority)(numberof(pthreadPriorityMapping) - 1);
	}

	struct sched_param param;
	param.sched_priority = pthreadPriorityMapping[pr];
	int ret = pthread_setschedparam(thread, SCHED_FIFO, &param);
	if (ret) fprintf(stderr, "pthread failed with error %d\n", ret);
	

/*
pthread_t tid;
int ret;
struct sched_param param;
int priority;
/* sched_priority will be the priority of the thread *
sched_param.sched_priority = priority;
/* only supported policy, others will result in ENOTSUP *

policy = SCHED_OTHER;
/* scheduling parameters of target thread *
ret = pthread_setschedparam(tid, policy, &param);

*/			
}

bool Thread::wait() {
	//not done yet
	return false;
}

#endif


//by default accept/run a callback and return its value

int Thread::run() {
	//run a callback if we were given one
	if (callback) return callback();
	return 0;
}

bool Thread::isDone() {
	return !thread;
}

void Thread::sleep(long ms) {
#ifdef WIN32
	Sleep(ms);
#else
#endif
}