#ifdef WIN32

#include <windows.h>
#include <assert.h>
#include "system/mutex.h"

class WinMutex : public IMutex {
public:
	CRITICAL_SECTION cs;
	bool initd;

	WinMutex(bool init = true);
	virtual ~WinMutex();
	
	virtual bool init();
	virtual void shutdown();
	
	virtual void lock();
	virtual void unlock();
};

WinMutex::WinMutex(bool initOnCtor) : initd(false) {
	if (initOnCtor) assert(init());
}

WinMutex::~WinMutex() { shutdown(); }

bool WinMutex::init() {
	InitializeCriticalSection(&cs);	//return code?
	initd = true;
	return true;
}

void WinMutex::shutdown() {
	if (initd) DeleteCriticalSection(&cs);
	initd = false;
}

void WinMutex::lock() {
	EnterCriticalSection(&cs);
}

void WinMutex::unlock() {
	LeaveCriticalSection(&cs);
}

IMutex *MutexFactory::create(bool init) {
	return new WinMutex(init);
}

#endif
